'' =================================================================================================
''
''   File....... jm_stepper.spin
''   Purpose.... Background driver for unipolor or bipolar stepper
''               * requires one cog for motor driver
''   Author..... Jon "JonnyMac" McPhalen
''               Copyright (C) 2017 Jon McPhalen
''               -- see below for terms of use
''   E-mail..... jon@jonmcphalen.com
''   Started....
''   Updated.... 19 JUL 2017
''
'' =================================================================================================


con { fixed io pins }                                                  
                                                                 
  RX1  = 31  { I }                                              ' serial / programming
  TX1  = 30  { O }                                            
                                                                 
  SDA1 = 29  { I/O }                                            ' i2c / eeprom
  SCL1 = 28  { I/O } 


con

  #true, ON, OFF                                                ' digital control   

  #0, M_WAVE, M_FULL, M_HALF                                    ' modes
  

var

  long  cog
  long  stack[64]

  long  ms1tix                                                  ' ticks in 1ms

  word  revsteps                                                ' steps in one revolution
  byte  apin                                                    ' A
  byte  bnpin                                                   ' /B
  long  epin                                                    ' enable pin (-1 if not used)
  word  p_table                                                 ' pointer to step table

  long  nsteps                                                  ' # steps to run
  byte  stpdir                                                  ' step direction (1 fwd, -1 rev)
  long  stptix                                                  ' step delay in system ticks      
  byte  stpidx                                                  ' step index

  byte  eflag                                                   ' enable flag
   
                                                                 
pub null

  ' This is not a top-level object
  

pub start(_revsteps, _apin, _epin, mode)

'' Start stepper motor control cog
'' -- revsteps are # of full steps per revolution
'' -- _apin if for phase A control
''    * pins ascend: A B /A /B
'' -- _epin is the enable pin for L293D interface (-1 if not used)
'' -- mode is M_WAVE, M_FULL, or M_HALF

  ms1tix   := clkfreq / 1000                                    ' ticks in 1ms

  revsteps := _revsteps                                         ' steps / revolution
  apin     := _apin                                             ' A control
  bnpin    := _apin + 3                                         ' /B control
  epin     := _epin                                             ' enable for L293D

  dira[bnpin..apin] := %0000                                    ' clear pins in master cog
  if (epin => 0)
    dira[epin] := 0
  
  case mode                                                     ' point to steps table
    M_WAVE : p_table := @WaveSteps
    M_HALF : p_table := @HalfSteps
    other  : p_table := @FullSteps                              ' default mode

  if (mode == M_HALF)
    revsteps <<= 1                                              ' double steps for half-step mode

  cog := cognew(stepper, @stack) + 1                            ' start the motor cog

  if (cog)
    enable(true)

  return cog


pub stop

'' Stops steper motor control cog

  if (cog)
    cogstop(cog - 1)
    cog := 0

  eflag := false


pub steps_per_rev

'' Returns # of steps to rotate 360 degrees

  return revsteps
    

pub enable(state)

'' Set eflag to true or false
'' -- when false, motor stops and outputs released

  eflag := (state <> false)

  
pub steps(n, speed)

'' Run n steps
'' -- negative steps for reverse
'' -- speed is step timing in 0.1ms units

  if (n == 0)
    return

  stptix := speed * ms1tix / 10                                 ' convert to system ticks

  if (n > 0)                    
    stpdir := 1                                                 ' forward   
  else 
    stpdir := -1                                                ' reverse   
    if (n == negx)                                              ' fix negx                                              
      n := posx

  stpidx := (stpidx + stpdir) & %111                            ' set 1st step of move
  nsteps := ||n                                                 ' set step count


pub timed_steps(n, duration)

'' Run n steps in duration
'' -- negative steps for reverse
'' -- duration is move timing in 0.1ms units

  steps(n, duration / ||n)


pub rpm_steps(n, rpm)

'' Run n steps at speed defined by RPM
'' -- negative steps for reverse

  steps(n, 60_000_0 / (rpm * revsteps))                         ' convert rpm to timing
  

pub degrees(d, speed) | count

'' Rotate d degrees 
'' -- negative d for reverse 
'' -- speed is step timing in 0.1ms units

  steps(d * revsteps / 360, speed)                              ' convert degrees to steps


pub timed_degrees(d, duration)

'' Rotate d degrees 
'' -- negative d for reverse
'' -- duration is move timing in 0.1ms units

  timed_steps(d * revsteps / 360, duration)


pub rpm_degrees(d, rpm)

'' Rotate d degrees at speed defined by rpm
'' -- negative d for reverse 

  steps(d * revsteps / 360, 60_000_0 / (rpm * revsteps))
  

pub halt

'' Stops motor before steps run out

  nsteps := 0
  

pub running

'' Return true when motor is running

  return (nsteps > 0)


dat { step tables }

' Step definition bits are (msb to lsb): /B /A B A

  WaveSteps     byte    %0001, %0010, %0100, %1000, %0001, %0010, %0100, %1000
  FullSteps     byte    %0011, %0110, %1100, %1001, %0011, %0110, %1100, %1001
  HalfSteps     byte    %0001, %0011, %0010, %0110, %0100, %1100, %1000, %1001
  

pri stepper | t

  dira[bnpin..apin] := %1111                                    ' make phase pins outputs
  if (epin => 0)                                                ' using enable pin?                                               '
    dira[epin] := 1                                             ' make output

  repeat                                                        ' motor loop
    motor_hold                                                  ' wait for enabled and steps
    t := cnt                                                    ' sync timing
    repeat while (nsteps > 0)                                   ' run all steps
      outa[bnpin..apin] := byte[p_table][stpidx]                ' output next step
      if (--nsteps)                                             ' if steps left
        waitcnt(t += stptix)                                    '  finish timing   
        stpidx := (stpidx + stpdir) & %111                      '  update step index      


pri motor_hold

'' Waits enabled and positive count in nsteps

  repeat
    if (eflag == false)                                         ' if not enabled           
      if (epin => 0)                                                                     
        outa[epin] := 0                                         ' disable l293d                                    
      else                                                                               
        outa[bnpin..apin] := %0000                              ' float open-drain outputs                                                      
      nsteps := 0                                               ' kill step count               
    else                                                                                 
      if (epin => 0)                                                                     
        outa[epin] := 1                                         ' enable l293d      
      else                                                                               
        outa[bnpin..apin] := byte[p_table][stpidx]              ' enable o.d. outputs

    if ((eflag) and (nsteps > 0))                               ' return when enabled with steps                                    
      quit

       
dat { license }

{{

  Terms of Use: MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy of this
  software and associated documentation files (the "Software"), to deal in the Software
  without restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be included in all copies
  or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
  PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 

}}      